/// \file example5.cc Example with the option of saving the particle configuration
/// to a file, and using saved configurations as input files.
///
/// Usage:
/// 
/// \code
///     ./example5 [ -fixed-order [ -order O ] | -all-order [ -non-recursive ] [ -t T ] ] 
///                [ -q | -g ] [ -seed S ] [ -nev N ] [ -eps E ] [ -nbin B ]
///                [ -fcut F ] [ -nfilt M ] [ -nsteps NS ] [ -dt DT ] [ -output FN ]
///                [ -save-config CN | -load-config CN ]
///                [ -ca CA ] [ -cf CF ] [ -tr TR] [ -nf NF ]
///               
/// \endcode
///
/// The code can be run to give fixed order results (-fixed-order),
/// specifically the coefficient of t^O, or all-order results
/// (-all-order) for a given value of t=T.
///
/// You can specify whether you start with a quark (default) or a
/// gluon with the -q and -g options.
///
/// Technical parameters (see default values in Input class):
///
/// - -non-recursive  specifies an alternative all-order implementation
/// - -nev N          indicates that the program should generate N events (default 10^4)
/// - -eps E          specifies the value of the cutoff: E < z < 1-E.
/// - -seed S         specifies the random seed (default 1)
/// - -nbin B         indicates the number of bins for histograms
/// - -fcut F         specifies the treshold value F < z for trimming
/// - -nfilt M        specifies the number of hardest subjets for filtering
/// - -nsteps NS      specifies the number of values t to compute
/// - -dt DT          specifies the increment in t at each step
/// - -output FN      indicates the name of output file (default is to std::cout)
/// - -save-config CN indicates the name of the configuration file where output will be written
/// - -load-config CN indicates the name of an existing configuration
/// - -ca CA          specifies the value of CA in pgg
/// - -cf CF          specifies the value of CF in pqq
/// - -tr TR          specifies the value of TR in pqg
/// - -nf NF          specifies the value of NF in pqg
///

#include "Input.hh"
#include "NLOHist.hh"
#include "AverageAndError.hh"
#include "Observable.hh"
#include "Output.hh"
#include "Generator.hh"
#include "AllOrderGeneratorFlat.hh"
#include "FixedOrderGenerator.hh"
#include "AllOrderGeneratorRecursive.hh"
#include <iostream>
#include <fstream>

using namespace std;

int main(int argc, char ** argv) {
  Input in(argc,argv);
  in.assert_all_options_used();
  cout << "# " << in.command << "\n# " << in.time << endl;
  Output output(in.filename,in.command,in.time);

  Generator * generator;
  if (in.fixedorder) {
    in.nsteps=1;
    generator = new FixedOrderGenerator(in.seed,in.flav,in.order,
					in.ca,in.cf,in.tr,in.nf);
  }
  else if (in.nonrecursive) {
    generator = new AllOrderGeneratorFlat(in.seed,in.flav,in.eps,in.t,
					  in.ca,in.cf,in.tr,in.nf);
  }
  else {
    generator = new AllOrderGeneratorRecursive(in.seed,in.flav,in.eps,in.t,
					       in.ca,in.cf,in.tr,in.nf);
  }
  std::fstream configstream;
  if (in.saveconfig) {
    configstream.open(in.configname.c_str(), std::fstream::out);
    configstream.precision(18);
  }
  if (in.loadconfig) {
    configstream.open(in.configname.c_str(), std::fstream::in);
    in.nev=1;
    std::cout << "reading file \n\n";
  }
  map<string,AverageAndError> averages;

  // declare vector NLOHist, with histos[i][k] for t=t+i*dt and
  // k=0 for hardest, k=1 for inclusive spectrum
  vector<NLOHist> histos(2);
  histos[0].declare("x max",0.0,1.+1e-9,in.nbin);
  histos[1].declare("x inclusive",0.0,1.+1e-9,in.nbin);
  
  Parton::Flavour changeflav;
  if (in.flav==Parton::q) changeflav=Parton::g;
  if (in.flav==Parton::g) changeflav=Parton::q;
  FlavHardest changeflavhardest(changeflav);
  Filter filter(in.nfilt);
  Trim trim(in.fcut);

  // Loop over events
  for (int iev = 0; iev < in.nev; iev++) {
    // events may come in "groups" (like for a NLO program);
    vector<Event> events;
    if (!in.loadconfig) {
      // generate the next group
      generator->next_event_group();
      // get a reference to the next group of events
      events = generator->events();
    } else if (configstream.is_open()){
      events = in.next_event_group(configstream);
    }
     
    // calculate observables
    averages["delta z"]   += average_over_group(events,deltaxmax);
    averages["log z"]     += average_over_group(events,logxmax);
    averages["prob flavour change"]
      += average_over_group(events,changeflavhardest);
    averages["z^6"]       += average_over_group(events,xmax6);
    averages["trimmed z"] += average_over_group(events,trim);
    averages["filtered z"]+= average_over_group(events,filter);
    add_group_to_histogram(histos[0],events,xmax);
    add_group_to_histogram(histos[1],events,xinclusive);      

    for(vector<NLOHist>::iterator it=histos.begin();
	it != histos.end(); ++it)
      it->collate_event();
    if (in.saveconfig)
      output.write(configstream,events);
    if (in.loadconfig && configstream.is_open()) 
      if (!in.file_end(configstream)) ++in.nev;
    // Write results when writing_step(iev)==true
    if(output.writing_step(iev)) {
      output.write(averages,histos,iev);
      cout << "# Currently at " << iev << " out of "
	   << in.nev << " events." << endl;
    }
    
    // Reset to initial configuration for next event
    if (!in.loadconfig)
      generator->reset();
  }
  if (in.saveconfig) configstream.close();
  output.write(averages,histos,in.nev);
  cout << "# Simulation completed" << endl;
}

